# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2009 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://genshi.edgewall.org/wiki/License.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://genshi.edgewall.org/log/.

"""Support for constructing markup streams from files, strings, or other
sources.
"""

from itertools import chain
import codecs
import htmlentitydefs as entities
import HTMLParser as html
from xml.parsers import expat

from genshi.core import Attrs, QName, Stream, stripentities
from genshi.core import START, END, XML_DECL, DOCTYPE, TEXT, START_NS, \
                        END_NS, START_CDATA, END_CDATA, PI, COMMENT
from genshi.compat import StringIO, BytesIO


__all__ = ['ET', 'ParseError', 'XMLParser', 'XML', 'HTMLParser', 'HTML']
__docformat__ = 'restructuredtext en'


def ET(element):
    """Convert a given ElementTree element to a markup stream.
    
    :param element: an ElementTree element
    :return: a markup stream
    """
    tag_name = QName(element.tag.lstrip('{'))
    attrs = Attrs([(QName(attr.lstrip('{')), value)
                   for attr, value in element.items()])

    yield START, (tag_name, attrs), (None, -1, -1)
    if element.text:
        yield TEXT, element.text, (None, -1, -1)
    for child in element.getchildren():
        for item in ET(child):
            yield item
    yield END, tag_name, (None, -1, -1)
    if element.tail:
        yield TEXT, element.tail, (None, -1, -1)


class ParseError(Exception):
    """Exception raised when fatal syntax errors are found in the input being
    parsed.
    """

    def __init__(self, message, filename=None, lineno=-1, offset=-1):
        """Exception initializer.
        
        :param message: the error message from the parser
        :param filename: the path to the file that was parsed
        :param lineno: the number of the line on which the error was encountered
        :param offset: the column number where the error was encountered
        """
        self.msg = message
        if filename:
            message += ', in ' + filename
        Exception.__init__(self, message)
        self.filename = filename or '<string>'
        self.lineno = lineno
        self.offset = offset


class XMLParser(object):
    """Generator-based XML parser based on roughly equivalent code in
    Kid/ElementTree.
    
    The parsing is initiated by iterating over the parser object:
    
    >>> parser = XMLParser(StringIO('<root id="2"><child>Foo</child></root>'))
    >>> for kind, data, pos in parser:
    ...     print('%s %s' % (kind, data))
    START (QName('root'), Attrs([(QName('id'), u'2')]))
    START (QName('child'), Attrs())
    TEXT Foo
    END child
    END root
    """

    _entitydefs = ['<!ENTITY %s "&#%d;">' % (name, value) for name, value in
                   entities.name2codepoint.items()]
    _external_dtd = u'\n'.join(_entitydefs).encode('utf-8')

    def __init__(self, source, filename=None, encoding=None):
        """Initialize the parser for the given XML input.
        
        :param source: the XML text as a file-like object
        :param filename: the name of the file, if appropriate
        :param encoding: the encoding of the file; if not specified, the
                         encoding is assumed to be ASCII, UTF-8, or UTF-16, or
                         whatever the encoding specified in the XML declaration
                         (if any)
        """
        self.source = source
        self.filename = filename

        # Setup the Expat parser
        parser = expat.ParserCreate(encoding, '}')
        parser.buffer_text = True
        # Python 3 does not have returns_unicode
        if hasattr(parser, 'returns_unicode'):
            parser.returns_unicode = True
        parser.ordered_attributes = True

        parser.StartElementHandler = self._handle_start
        parser.EndElementHandler = self._handle_end
        parser.CharacterDataHandler = self._handle_data
        parser.StartDoctypeDeclHandler = self._handle_doctype
        parser.StartNamespaceDeclHandler = self._handle_start_ns
        parser.EndNamespaceDeclHandler = self._handle_end_ns
        parser.StartCdataSectionHandler = self._handle_start_cdata
        parser.EndCdataSectionHandler = self._handle_end_cdata
        parser.ProcessingInstructionHandler = self._handle_pi
        parser.XmlDeclHandler = self._handle_xml_decl
        parser.CommentHandler = self._handle_comment

        # Tell Expat that we'll handle non-XML entities ourselves
        # (in _handle_other)
        parser.DefaultHandler = self._handle_other
        parser.SetParamEntityParsing(expat.XML_PARAM_ENTITY_PARSING_ALWAYS)
        parser.UseForeignDTD()
        parser.ExternalEntityRefHandler = self._build_foreign

        self.expat = parser
        self._queue = []

    def parse(self):
        """Generator that parses the XML source, yielding markup events.
        
        :return: a markup event stream
        :raises ParseError: if the XML text is not well formed
        """
        def _generate():
            try:
                bufsize = 4 * 1024 # 4K
                done = False
                while 1:
                    while not done and len(self._queue) == 0:
                        data = self.source.read(bufsize)
                        if not data: # end of data
                            if hasattr(self, 'expat'):
                                self.expat.Parse('', True)
                                del self.expat # get rid of circular references
                            done = True
                        else:
                            if isinstance(data, unicode):
                                data = data.encode('utf-8')
                            self.expat.Parse(data, False)
                    for event in self._queue:
                        yield event
                    self._queue = []
                    if done:
                        break
            except expat.ExpatError, e:
                msg = str(e)
                raise ParseError(msg, self.filename, e.lineno, e.offset)
        return Stream(_generate()).filter(_coalesce)

    def __iter__(self):
        return iter(self.parse())

    def _build_foreign(self, context, base, sysid, pubid):
        parser = self.expat.ExternalEntityParserCreate(context)
        parser.ParseFile(BytesIO(self._external_dtd))
        return 1

    def _enqueue(self, kind, data=None, pos=None):
        if pos is None:
            pos = self._getpos()
        if kind is TEXT:
            # Expat reports the *end* of the text event as current position. We
            # try to fix that up here as much as possible. Unfortunately, the
            # offset is only valid for single-line text. For multi-line text,
            # it is apparently not possible to determine at what offset it
            # started
            if '\n' in data:
                lines = data.splitlines()
                lineno = pos[1] - len(lines) + 1
                offset = -1
            else:
                lineno = pos[1]
                offset = pos[2] - len(data)
            pos = (pos[0], lineno, offset)
        self._queue.append((kind, data, pos))

    def _getpos_unknown(self):
        return (self.filename, -1, -1)

    def _getpos(self):
        return (self.filename, self.expat.CurrentLineNumber,
                self.expat.CurrentColumnNumber)

    def _handle_start(self, tag, attrib):
        attrs = Attrs([(QName(name), value) for name, value in
                       zip(*[iter(attrib)] * 2)])
        self._enqueue(START, (QName(tag), attrs))

    def _handle_end(self, tag):
        self._enqueue(END, QName(tag))

    def _handle_data(self, text):
        self._enqueue(TEXT, text)

    def _handle_xml_decl(self, version, encoding, standalone):
        self._enqueue(XML_DECL, (version, encoding, standalone))

    def _handle_doctype(self, name, sysid, pubid, has_internal_subset):
        self._enqueue(DOCTYPE, (name, pubid, sysid))

    def _handle_start_ns(self, prefix, uri):
        self._enqueue(START_NS, (prefix or '', uri))

    def _handle_end_ns(self, prefix):
        self._enqueue(END_NS, prefix or '')

    def _handle_start_cdata(self):
        self._enqueue(START_CDATA)

    def _handle_end_cdata(self):
        self._enqueue(END_CDATA)

    def _handle_pi(self, target, data):
        self._enqueue(PI, (target, data))

    def _handle_comment(self, text):
        self._enqueue(COMMENT, text)

    def _handle_other(self, text):
        if text.startswith('&'):
            # deal with undefined entities
            try:
                text = unichr(entities.name2codepoint[text[1:-1]])
                self._enqueue(TEXT, text)
            except KeyError:
                filename, lineno, offset = self._getpos()
                error = expat.error('undefined entity "%s": line %d, column %d'
                                    % (text, lineno, offset))
                error.code = expat.errors.XML_ERROR_UNDEFINED_ENTITY
                error.lineno = lineno
                error.offset = offset
                raise error


def XML(text):
    """Parse the given XML source and return a markup stream.
    
    Unlike with `XMLParser`, the returned stream is reusable, meaning it can be
    iterated over multiple times:
    
    >>> xml = XML('<doc><elem>Foo</elem><elem>Bar</elem></doc>')
    >>> print(xml)
    <doc><elem>Foo</elem><elem>Bar</elem></doc>
    >>> print(xml.select('elem'))
    <elem>Foo</elem><elem>Bar</elem>
    >>> print(xml.select('elem/text()'))
    FooBar
    
    :param text: the XML source
    :return: the parsed XML event stream
    :raises ParseError: if the XML text is not well-formed
    """
    return Stream(list(XMLParser(StringIO(text))))


class HTMLParser(html.HTMLParser, object):
    """Parser for HTML input based on the Python `HTMLParser` module.
    
    This class provides the same interface for generating stream events as
    `XMLParser`, and attempts to automatically balance tags.
    
    The parsing is initiated by iterating over the parser object:
    
    >>> parser = HTMLParser(BytesIO(u'<UL compact><LI>Foo</UL>'.encode('utf-8')), encoding='utf-8')
    >>> for kind, data, pos in parser:
    ...     print('%s %s' % (kind, data))
    START (QName('ul'), Attrs([(QName('compact'), u'compact')]))
    START (QName('li'), Attrs())
    TEXT Foo
    END li
    END ul
    """

    _EMPTY_ELEMS = frozenset(['area', 'base', 'basefont', 'br', 'col', 'frame',
                              'hr', 'img', 'input', 'isindex', 'link', 'meta',
                              'param'])

    def __init__(self, source, filename=None, encoding=None):
        """Initialize the parser for the given HTML input.
        
        :param source: the HTML text as a file-like object
        :param filename: the name of the file, if known
        :param filename: encoding of the file; ignored if the input is unicode
        """
        html.HTMLParser.__init__(self)
        self.source = source
        self.filename = filename
        self.encoding = encoding
        self._queue = []
        self._open_tags = []

    def parse(self):
        """Generator that parses the HTML source, yielding markup events.
        
        :return: a markup event stream
        :raises ParseError: if the HTML text is not well formed
        """
        def _generate():
            if self.encoding:
                reader = codecs.getreader(self.encoding)
                source = reader(self.source)
            else:
                source = self.source
            try:
                bufsize = 4 * 1024 # 4K
                done = False
                while 1:
                    while not done and len(self._queue) == 0:
                        data = source.read(bufsize)
                        if not data: # end of data
                            self.close()
                            done = True
                        else:
                            if not isinstance(data, unicode):
                                raise UnicodeError("source returned bytes, but no encoding specified")
                            self.feed(data)
                    for kind, data, pos in self._queue:
                        yield kind, data, pos
                    self._queue = []
                    if done:
                        open_tags = self._open_tags
                        open_tags.reverse()
                        for tag in open_tags:
                            yield END, QName(tag), pos
                        break
            except html.HTMLParseError, e:
                msg = '%s: line %d, column %d' % (e.msg, e.lineno, e.offset)
                raise ParseError(msg, self.filename, e.lineno, e.offset)
        return Stream(_generate()).filter(_coalesce)

    def __iter__(self):
        return iter(self.parse())

    def _enqueue(self, kind, data, pos=None):
        if pos is None:
            pos = self._getpos()
        self._queue.append((kind, data, pos))

    def _getpos(self):
        lineno, column = self.getpos()
        return (self.filename, lineno, column)

    def handle_starttag(self, tag, attrib):
        fixed_attrib = []
        for name, value in attrib: # Fixup minimized attributes
            if value is None:
                value = name
            fixed_attrib.append((QName(name), stripentities(value)))

        self._enqueue(START, (QName(tag), Attrs(fixed_attrib)))
        if tag in self._EMPTY_ELEMS:
            self._enqueue(END, QName(tag))
        else:
            self._open_tags.append(tag)

    def handle_endtag(self, tag):
        if tag not in self._EMPTY_ELEMS:
            while self._open_tags:
                open_tag = self._open_tags.pop()
                self._enqueue(END, QName(open_tag))
                if open_tag.lower() == tag.lower():
                    break

    def handle_data(self, text):
        self._enqueue(TEXT, text)

    def handle_charref(self, name):
        if name.lower().startswith('x'):
            text = unichr(int(name[1:], 16))
        else:
            text = unichr(int(name))
        self._enqueue(TEXT, text)

    def handle_entityref(self, name):
        try:
            text = unichr(entities.name2codepoint[name])
        except KeyError:
            text = '&%s;' % name
        self._enqueue(TEXT, text)

    def handle_pi(self, data):
        if data.endswith('?'):
            data = data[:-1]
        try:
            target, data = data.split(None, 1)
        except ValueError:
            # PI with no data
            target = data
            data = ''
        self._enqueue(PI, (target.strip(), data.strip()))

    def handle_comment(self, text):
        self._enqueue(COMMENT, text)


def HTML(text, encoding=None):
    """Parse the given HTML source and return a markup stream.
    
    Unlike with `HTMLParser`, the returned stream is reusable, meaning it can be
    iterated over multiple times:
    
    >>> html = HTML('<body><h1>Foo</h1></body>', encoding='utf-8')
    >>> print(html)
    <body><h1>Foo</h1></body>
    >>> print(html.select('h1'))
    <h1>Foo</h1>
    >>> print(html.select('h1/text()'))
    Foo
    
    :param text: the HTML source
    :return: the parsed XML event stream
    :raises ParseError: if the HTML text is not well-formed, and error recovery
                        fails
    """
    if isinstance(text, unicode):
        # If it's unicode text the encoding should be set to None.
        # The option to pass in an incorrect encoding is for ease
        # of writing doctests that work in both Python 2.x and 3.x.
        return Stream(list(HTMLParser(StringIO(text), encoding=None)))
    return Stream(list(HTMLParser(BytesIO(text), encoding=encoding)))


def _coalesce(stream):
    """Coalesces adjacent TEXT events into a single event."""
    textbuf = []
    textpos = None
    for kind, data, pos in chain(stream, [(None, None, None)]):
        if kind is TEXT:
            textbuf.append(data)
            if textpos is None:
                textpos = pos
        else:
            if textbuf:
                yield TEXT, ''.join(textbuf), textpos
                del textbuf[:]
                textpos = None
            if kind:
                yield kind, data, pos



# -*- coding: utf-8 -*-
#
# Copyright (C) 2007 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://genshi.edgewall.org/wiki/License.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://genshi.edgewall.org/log/.

import doctest
from pprint import pprint
import unittest

from genshi import HTML
from genshi.builder import Element
from genshi.core import START, END, TEXT, QName, Attrs
from genshi.filters.transform import Transformer, StreamBuffer, ENTER, EXIT, \
                                     OUTSIDE, INSIDE, ATTR, BREAK
import genshi.filters.transform


FOO = '<root>ROOT<foo name="foo">FOO</foo></root>'
FOOBAR = '<root>ROOT<foo name="foo" size="100">FOO</foo><bar name="bar">BAR</bar></root>'


def _simplify(stream, with_attrs=False):
    """Simplify a marked stream."""
    def _generate():
        for mark, (kind, data, pos) in stream:
            if kind is START:
                if with_attrs:
                    data = (unicode(data[0]), dict((unicode(k), v)
                                                   for k, v in data[1]))
                else:
                    data = unicode(data[0])
            elif kind is END:
                data = unicode(data)
            elif kind is ATTR:
                kind = ATTR
                data = dict((unicode(k), v) for k, v in data[1])
            yield mark, kind, data
    return list(_generate())


def _transform(html, transformer, with_attrs=False):
    """Apply transformation returning simplified marked stream."""
    if isinstance(html, basestring) and not isinstance(html, unicode):
        html = HTML(html, encoding='utf-8')
    elif isinstance(html, unicode):
        html = HTML(html, encoding='utf-8')
    stream = transformer(html, keep_marks=True)
    return _simplify(stream, with_attrs)


class SelectTest(unittest.TestCase):
    """Test .select()"""
    def _select(self, select):
        html = HTML(FOOBAR, encoding='utf-8')
        if isinstance(select, basestring):
            select = [select]
        transformer = Transformer(select[0])
        for sel in select[1:]:
            transformer = transformer.select(sel)
        return _transform(html, transformer)

    def test_select_single_element(self):
        self.assertEqual(
            self._select('foo'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')],
            )

    def test_select_context(self):
        self.assertEqual(
            self._select('.'),
            [(ENTER, START, u'root'),
             (INSIDE, TEXT, u'ROOT'),
             (INSIDE, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (INSIDE, END, u'foo'),
             (INSIDE, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (INSIDE, END, u'bar'),
             (EXIT, END, u'root')]
            )

    def test_select_inside_select(self):
        self.assertEqual(
            self._select(['.', 'foo']),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')],
            )

    def test_select_text(self):
        self.assertEqual(
            self._select('*/text()'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, START, u'foo'),
             (OUTSIDE, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, START, u'bar'),
             (OUTSIDE, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')],
            )

    def test_select_attr(self):
        self.assertEqual(
            self._select('foo/@name'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ATTR, ATTR, {'name': u'foo'}),
             (None, START, u'foo'),
             (None, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_select_text_context(self):
        self.assertEqual(
            list(Transformer('.')(HTML(u'foo'), keep_marks=True)),
            [('OUTSIDE', ('TEXT', u'foo', (None, 1, 0)))],
            )


class InvertTest(unittest.TestCase):
    def _invert(self, select):
        return _transform(FOO, Transformer(select).invert())

    def test_invert_element(self):
        self.assertEqual(
            self._invert('foo'),
            [(OUTSIDE, START, u'root'),
             (OUTSIDE, TEXT, u'ROOT'),
             (None, START, u'foo'),
             (None, TEXT, u'FOO'),
             (None, END, u'foo'),
             (OUTSIDE, END, u'root')]
            )

    def test_invert_inverted_element(self):
        self.assertEqual(
            _transform(FOO, Transformer('foo').invert().invert()),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (OUTSIDE, START, u'foo'),
             (OUTSIDE, TEXT, u'FOO'),
             (OUTSIDE, END, u'foo'),
             (None, END, u'root')]
            )

    def test_invert_text(self):
        self.assertEqual(
            self._invert('foo/text()'),
            [(OUTSIDE, START, u'root'),
             (OUTSIDE, TEXT, u'ROOT'),
             (OUTSIDE, START, u'foo'),
             (None, TEXT, u'FOO'),
             (OUTSIDE, END, u'foo'),
             (OUTSIDE, END, u'root')]
            )

    def test_invert_attribute(self):
        self.assertEqual(
            self._invert('foo/@name'),
            [(OUTSIDE, START, u'root'),
             (OUTSIDE, TEXT, u'ROOT'),
             (None, ATTR, {'name': u'foo'}),
             (OUTSIDE, START, u'foo'),
             (OUTSIDE, TEXT, u'FOO'),
             (OUTSIDE, END, u'foo'),
             (OUTSIDE, END, u'root')]
            )

    def test_invert_context(self):
        self.assertEqual(
            self._invert('.'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, START, u'foo'),
             (None, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, END, u'root')]
            )

    def test_invert_text_context(self):
        self.assertEqual(
            _simplify(Transformer('.').invert()(HTML(u'foo'), keep_marks=True)),
            [(None, 'TEXT', u'foo')],
            )



class EndTest(unittest.TestCase):
    def test_end(self):
        stream = _transform(FOO, Transformer('foo').end())
        self.assertEqual(
            stream,
            [(OUTSIDE, START, u'root'),
             (OUTSIDE, TEXT, u'ROOT'),
             (OUTSIDE, START, u'foo'),
             (OUTSIDE, TEXT, u'FOO'),
             (OUTSIDE, END, u'foo'),
             (OUTSIDE, END, u'root')]
            )


class EmptyTest(unittest.TestCase):
    def _empty(self, select):
        return _transform(FOO, Transformer(select).empty())

    def test_empty_element(self):
        self.assertEqual(
            self._empty('foo'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (EXIT, END, u'foo'),
             (None, END, u'root')],
            )

    def test_empty_text(self):
        self.assertEqual(
            self._empty('foo/text()'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, START, u'foo'),
             (OUTSIDE, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, END, u'root')]
            )

    def test_empty_attr(self):
        self.assertEqual(
            self._empty('foo/@name'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ATTR, ATTR, {'name': u'foo'}),
             (None, START, u'foo'),
             (None, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, END, u'root')]
            )

    def test_empty_context(self):
        self.assertEqual(
            self._empty('.'),
            [(ENTER, START, u'root'),
             (EXIT, END, u'root')]
            )

    def test_empty_text_context(self):
        self.assertEqual(
            _simplify(Transformer('.')(HTML(u'foo'), keep_marks=True)),
            [(OUTSIDE, TEXT, u'foo')],
            )


class RemoveTest(unittest.TestCase):
    def _remove(self, select):
        return _transform(FOO, Transformer(select).remove())

    def test_remove_element(self):
        self.assertEqual(
            self._remove('foo|bar'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, END, u'root')]
            )

    def test_remove_text(self):
        self.assertEqual(
            self._remove('//text()'),
            [(None, START, u'root'),
             (None, START, u'foo'),
             (None, END, u'foo'),
             (None, END, u'root')]
            )

    def test_remove_attr(self):
        self.assertEqual(
            self._remove('foo/@name'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, START, u'foo'),
             (None, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, END, u'root')]
            )

    def test_remove_context(self):
        self.assertEqual(
            self._remove('.'),
            [],
            )

    def test_remove_text_context(self):
        self.assertEqual(
            _transform('foo', Transformer('.').remove()),
            [],
            )


class UnwrapText(unittest.TestCase):
    def _unwrap(self, select):
        return _transform(FOO, Transformer(select).unwrap())

    def test_unwrap_element(self):
        self.assertEqual(
            self._unwrap('foo'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (INSIDE, TEXT, u'FOO'),
             (None, END, u'root')]
            )

    def test_unwrap_text(self):
        self.assertEqual(
            self._unwrap('foo/text()'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, START, u'foo'),
             (OUTSIDE, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, END, u'root')]
            )

    def test_unwrap_attr(self):
        self.assertEqual(
            self._unwrap('foo/@name'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ATTR, ATTR, {'name': u'foo'}),
             (None, START, u'foo'),
             (None, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, END, u'root')]
            )

    def test_unwrap_adjacent(self):
        self.assertEqual(
            _transform(FOOBAR, Transformer('foo|bar').unwrap()),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (INSIDE, TEXT, u'FOO'),
             (INSIDE, TEXT, u'BAR'),
             (None, END, u'root')]
            )

    def test_unwrap_root(self):
        self.assertEqual(
            self._unwrap('.'),
            [(INSIDE, TEXT, u'ROOT'),
             (INSIDE, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (INSIDE, END, u'foo')]
            )

    def test_unwrap_text_root(self):
        self.assertEqual(
            _transform('foo', Transformer('.').unwrap()),
            [(OUTSIDE, TEXT, 'foo')],
            )


class WrapTest(unittest.TestCase):
    def _wrap(self, select, wrap='wrap'):
        return _transform(FOO, Transformer(select).wrap(wrap))

    def test_wrap_element(self):
        self.assertEqual(
            self._wrap('foo'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, START, u'wrap'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, END, u'wrap'),
             (None, END, u'root')]
            )

    def test_wrap_adjacent_elements(self):
        self.assertEqual(
            _transform(FOOBAR, Transformer('foo|bar').wrap('wrap')),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, START, u'wrap'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, END, u'wrap'),
             (None, START, u'wrap'),
             (ENTER, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (EXIT, END, u'bar'),
             (None, END, u'wrap'),
             (None, END, u'root')]
            )

    def test_wrap_text(self):
        self.assertEqual(
            self._wrap('foo/text()'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, START, u'foo'),
             (None, START, u'wrap'),
             (OUTSIDE, TEXT, u'FOO'),
             (None, END, u'wrap'),
             (None, END, u'foo'),
             (None, END, u'root')]
            )

    def test_wrap_root(self):
        self.assertEqual(
            self._wrap('.'),
            [(None, START, u'wrap'),
             (ENTER, START, u'root'),
             (INSIDE, TEXT, u'ROOT'),
             (INSIDE, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (INSIDE, END, u'foo'),
             (EXIT, END, u'root'),
             (None, END, u'wrap')]
            )

    def test_wrap_text_root(self):
        self.assertEqual(
            _transform('foo', Transformer('.').wrap('wrap')),
            [(None, START, u'wrap'),
             (OUTSIDE, TEXT, u'foo'),
             (None, END, u'wrap')],
            )

    def test_wrap_with_element(self):
        element = Element('a', href='http://localhost')
        self.assertEqual(
            _transform('foo', Transformer('.').wrap(element), with_attrs=True),
            [(None, START, (u'a', {u'href': u'http://localhost'})),
             (OUTSIDE, TEXT, u'foo'),
             (None, END, u'a')]
            )


class FilterTest(unittest.TestCase):
    def _filter(self, select, html=FOOBAR):
        """Returns a list of lists of filtered elements."""
        output = []
        def filtered(stream):
            interval = []
            output.append(interval)
            for event in stream:
                interval.append(event)
                yield event
        _transform(html, Transformer(select).filter(filtered))
        simplified = []
        for sub in output:
            simplified.append(_simplify([(None, event) for event in sub]))
        return simplified

    def test_filter_element(self):
        self.assertEqual(
            self._filter('foo'),
            [[(None, START, u'foo'),
              (None, TEXT, u'FOO'),
              (None, END, u'foo')]]
            )

    def test_filter_adjacent_elements(self):
        self.assertEqual(
            self._filter('foo|bar'),
            [[(None, START, u'foo'),
              (None, TEXT, u'FOO'),
              (None, END, u'foo')],
             [(None, START, u'bar'),
              (None, TEXT, u'BAR'),
              (None, END, u'bar')]]
            )

    def test_filter_text(self):
        self.assertEqual(
            self._filter('*/text()'),
            [[(None, TEXT, u'FOO')],
             [(None, TEXT, u'BAR')]]
            )
    def test_filter_root(self):
        self.assertEqual(
            self._filter('.'),
            [[(None, START, u'root'),
              (None, TEXT, u'ROOT'),
              (None, START, u'foo'),
              (None, TEXT, u'FOO'),
              (None, END, u'foo'),
              (None, START, u'bar'),
              (None, TEXT, u'BAR'),
              (None, END, u'bar'),
              (None, END, u'root')]]
            )

    def test_filter_text_root(self):
        self.assertEqual(
            self._filter('.', 'foo'),
            [[(None, TEXT, u'foo')]])

    def test_filter_after_outside(self):
        stream = _transform(
            '<root>x</root>', Transformer('//root/text()').filter(lambda x: x))
        self.assertEqual(
            list(stream),
            [(None, START, u'root'),
             (OUTSIDE, TEXT, u'x'),
             (None, END, u'root')])


class MapTest(unittest.TestCase):
    def _map(self, select, kind=None):
        data = []
        def record(d):
            data.append(d)
            return d
        _transform(FOOBAR, Transformer(select).map(record, kind))
        return data

    def test_map_element(self):
        self.assertEqual(
            self._map('foo'),
            [(QName('foo'), Attrs([(QName('name'), u'foo'),
                                   (QName('size'), u'100')])),
             u'FOO',
             QName('foo')]
        )

    def test_map_with_text_kind(self):
        self.assertEqual(
            self._map('.', TEXT),
            [u'ROOT', u'FOO', u'BAR']
        )

    def test_map_with_root_and_end_kind(self):
        self.assertEqual(
            self._map('.', END),
            [QName('foo'), QName('bar'), QName('root')]
        )

    def test_map_with_attribute(self):
        self.assertEqual(
            self._map('foo/@name'),
            [(QName('foo@*'), Attrs([('name', u'foo')]))]
        )


class SubstituteTest(unittest.TestCase):
    def _substitute(self, select, pattern, replace):
        return _transform(FOOBAR, Transformer(select).substitute(pattern, replace))

    def test_substitute_foo(self):
        self.assertEqual(
            self._substitute('foo', 'FOO|BAR', 'FOOOOO'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOOOOO'),
             (EXIT, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_substitute_foobar_with_group(self):
        self.assertEqual(
            self._substitute('foo|bar', '(FOO|BAR)', r'(\1)'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'(FOO)'),
             (EXIT, END, u'foo'),
             (ENTER, START, u'bar'),
             (INSIDE, TEXT, u'(BAR)'),
             (EXIT, END, u'bar'),
             (None, END, u'root')]
            )


class RenameTest(unittest.TestCase):
    def _rename(self, select):
        return _transform(FOOBAR, Transformer(select).rename('foobar'))

    def test_rename_root(self):
        self.assertEqual(
            self._rename('.'),
            [(ENTER, START, u'foobar'),
             (INSIDE, TEXT, u'ROOT'),
             (INSIDE, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (INSIDE, END, u'foo'),
             (INSIDE, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (INSIDE, END, u'bar'),
             (EXIT, END, u'foobar')]
            )

    def test_rename_element(self):
        self.assertEqual(
            self._rename('foo|bar'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ENTER, START, u'foobar'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foobar'),
             (ENTER, START, u'foobar'),
             (INSIDE, TEXT, u'BAR'),
             (EXIT, END, u'foobar'),
             (None, END, u'root')]
            )

    def test_rename_text(self):
        self.assertEqual(
            self._rename('foo/text()'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, START, u'foo'),
             (OUTSIDE, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )


class ContentTestMixin(object):
    def _apply(self, select, content=None, html=FOOBAR):
        class Injector(object):
            count = 0

            def __iter__(self):
                self.count += 1
                return iter(HTML(u'CONTENT %i' % self.count))

        if isinstance(html, basestring) and not isinstance(html, unicode):
            html = HTML(html, encoding='utf-8')
        else:
            html = HTML(html)
        if content is None:
            content = Injector()
        elif isinstance(content, basestring):
            content = HTML(content)
        return _transform(html, getattr(Transformer(select), self.operation)
                                (content))


class ReplaceTest(unittest.TestCase, ContentTestMixin):
    operation = 'replace'

    def test_replace_element(self):
        self.assertEqual(
            self._apply('foo'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, TEXT, u'CONTENT 1'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_replace_text(self):
        self.assertEqual(
            self._apply('text()'),
            [(None, START, u'root'),
             (None, TEXT, u'CONTENT 1'),
             (None, START, u'foo'),
             (None, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_replace_context(self):
        self.assertEqual(
            self._apply('.'),
            [(None, TEXT, u'CONTENT 1')],
            )

    def test_replace_text_context(self):
        self.assertEqual(
            self._apply('.', html='foo'),
            [(None, TEXT, u'CONTENT 1')],
            )

    def test_replace_adjacent_elements(self):
        self.assertEqual(
            self._apply('*'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, TEXT, u'CONTENT 1'),
             (None, TEXT, u'CONTENT 2'),
             (None, END, u'root')],
            )

    def test_replace_all(self):
        self.assertEqual(
            self._apply('*|text()'),
            [(None, START, u'root'),
             (None, TEXT, u'CONTENT 1'),
             (None, TEXT, u'CONTENT 2'),
             (None, TEXT, u'CONTENT 3'),
             (None, END, u'root')],
            )

    def test_replace_with_callback(self):
        count = [0]
        def content():
            count[0] += 1
            yield '%2i.' % count[0]
        self.assertEqual(
            self._apply('*', content),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, TEXT, u' 1.'),
             (None, TEXT, u' 2.'),
             (None, END, u'root')]
            )


class BeforeTest(unittest.TestCase, ContentTestMixin):
    operation = 'before'

    def test_before_element(self):
        self.assertEqual(
            self._apply('foo'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, TEXT, u'CONTENT 1'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_before_text(self):
        self.assertEqual(
            self._apply('text()'),
            [(None, START, u'root'),
             (None, TEXT, u'CONTENT 1'),
             (OUTSIDE, TEXT, u'ROOT'),
             (None, START, u'foo'),
             (None, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_before_context(self):
        self.assertEqual(
            self._apply('.'),
            [(None, TEXT, u'CONTENT 1'),
             (ENTER, START, u'root'),
             (INSIDE, TEXT, u'ROOT'),
             (INSIDE, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (INSIDE, END, u'foo'),
             (INSIDE, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (INSIDE, END, u'bar'),
             (EXIT, END, u'root')]
            )

    def test_before_text_context(self):
        self.assertEqual(
            self._apply('.', html='foo'),
            [(None, TEXT, u'CONTENT 1'),
             (OUTSIDE, TEXT, u'foo')]
            )

    def test_before_adjacent_elements(self):
        self.assertEqual(
            self._apply('*'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (None, TEXT, u'CONTENT 1'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, TEXT, u'CONTENT 2'),
             (ENTER, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (EXIT, END, u'bar'),
             (None, END, u'root')]

            )

    def test_before_all(self):
        self.assertEqual(
            self._apply('*|text()'),
            [(None, START, u'root'),
             (None, TEXT, u'CONTENT 1'),
             (OUTSIDE, TEXT, u'ROOT'),
             (None, TEXT, u'CONTENT 2'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, TEXT, u'CONTENT 3'),
             (ENTER, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (EXIT, END, u'bar'),
             (None, END, u'root')]
            )

    def test_before_with_callback(self):
        count = [0]
        def content():
            count[0] += 1
            yield '%2i.' % count[0]
        self.assertEqual(
            self._apply('foo/text()', content),
            [(None, 'START', u'root'),
             (None, 'TEXT', u'ROOT'),
             (None, 'START', u'foo'),
             (None, 'TEXT', u' 1.'),
             ('OUTSIDE', 'TEXT', u'FOO'),
             (None, 'END', u'foo'),
             (None, 'START', u'bar'),
             (None, 'TEXT', u'BAR'),
             (None, 'END', u'bar'),
             (None, 'END', u'root')]
            )


class AfterTest(unittest.TestCase, ContentTestMixin):
    operation = 'after'

    def test_after_element(self):
        self.assertEqual(
            self._apply('foo'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, TEXT, u'CONTENT 1'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_after_text(self):
        self.assertEqual(
            self._apply('text()'),
            [(None, START, u'root'),
             (OUTSIDE, TEXT, u'ROOT'),
             (None, TEXT, u'CONTENT 1'),
             (None, START, u'foo'),
             (None, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_after_context(self):
        self.assertEqual(
            self._apply('.'),
            [(ENTER, START, u'root'),
             (INSIDE, TEXT, u'ROOT'),
             (INSIDE, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (INSIDE, END, u'foo'),
             (INSIDE, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (INSIDE, END, u'bar'),
             (EXIT, END, u'root'),
             (None, TEXT, u'CONTENT 1')]
            )

    def test_after_text_context(self):
        self.assertEqual(
            self._apply('.', html='foo'),
            [(OUTSIDE, TEXT, u'foo'),
             (None, TEXT, u'CONTENT 1')]
            )

    def test_after_adjacent_elements(self):
        self.assertEqual(
            self._apply('*'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, TEXT, u'CONTENT 1'),
             (ENTER, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (EXIT, END, u'bar'),
             (None, TEXT, u'CONTENT 2'),
             (None, END, u'root')]

            )

    def test_after_all(self):
        self.assertEqual(
            self._apply('*|text()'),
            [(None, START, u'root'),
             (OUTSIDE, TEXT, u'ROOT'),
             (None, TEXT, u'CONTENT 1'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, TEXT, u'CONTENT 2'),
             (ENTER, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (EXIT, END, u'bar'),
             (None, TEXT, u'CONTENT 3'),
             (None, END, u'root')]
            )

    def test_after_with_callback(self):
        count = [0]
        def content():
            count[0] += 1
            yield '%2i.' % count[0]
        self.assertEqual(
            self._apply('foo/text()', content),
            [(None, 'START', u'root'),
             (None, 'TEXT', u'ROOT'),
             (None, 'START', u'foo'),
             ('OUTSIDE', 'TEXT', u'FOO'),
             (None, 'TEXT', u' 1.'),
             (None, 'END', u'foo'),
             (None, 'START', u'bar'),
             (None, 'TEXT', u'BAR'),
             (None, 'END', u'bar'),
             (None, 'END', u'root')]
            )


class PrependTest(unittest.TestCase, ContentTestMixin):
    operation = 'prepend'

    def test_prepend_element(self):
        self.assertEqual(
            self._apply('foo'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (None, TEXT, u'CONTENT 1'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_prepend_text(self):
        self.assertEqual(
            self._apply('text()'),
            [(None, START, u'root'),
             (OUTSIDE, TEXT, u'ROOT'),
             (None, START, u'foo'),
             (None, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_prepend_context(self):
        self.assertEqual(
            self._apply('.'),
            [(ENTER, START, u'root'),
             (None, TEXT, u'CONTENT 1'),
             (INSIDE, TEXT, u'ROOT'),
             (INSIDE, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (INSIDE, END, u'foo'),
             (INSIDE, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (INSIDE, END, u'bar'),
             (EXIT, END, u'root')],
            )

    def test_prepend_text_context(self):
        self.assertEqual(
            self._apply('.', html='foo'),
            [(OUTSIDE, TEXT, u'foo')]
            )

    def test_prepend_adjacent_elements(self):
        self.assertEqual(
            self._apply('*'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (None, TEXT, u'CONTENT 1'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (ENTER, START, u'bar'),
             (None, TEXT, u'CONTENT 2'),
             (INSIDE, TEXT, u'BAR'),
             (EXIT, END, u'bar'),
             (None, END, u'root')]

            )

    def test_prepend_all(self):
        self.assertEqual(
            self._apply('*|text()'),
            [(None, START, u'root'),
             (OUTSIDE, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (None, TEXT, u'CONTENT 1'),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (ENTER, START, u'bar'),
             (None, TEXT, u'CONTENT 2'),
             (INSIDE, TEXT, u'BAR'),
             (EXIT, END, u'bar'),
             (None, END, u'root')]
            )

    def test_prepend_with_callback(self):
        count = [0]
        def content():
            count[0] += 1
            yield '%2i.' % count[0]
        self.assertEqual(
            self._apply('foo', content),
            [(None, 'START', u'root'),
             (None, 'TEXT', u'ROOT'),
             (ENTER, 'START', u'foo'),
             (None, 'TEXT', u' 1.'),
             (INSIDE, 'TEXT', u'FOO'),
             (EXIT, 'END', u'foo'),
             (None, 'START', u'bar'),
             (None, 'TEXT', u'BAR'),
             (None, 'END', u'bar'),
             (None, 'END', u'root')]
            )


class AppendTest(unittest.TestCase, ContentTestMixin):
    operation = 'append'

    def test_append_element(self):
        self.assertEqual(
            self._apply('foo'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (None, TEXT, u'CONTENT 1'),
             (EXIT, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_append_text(self):
        self.assertEqual(
            self._apply('text()'),
            [(None, START, u'root'),
             (OUTSIDE, TEXT, u'ROOT'),
             (None, START, u'foo'),
             (None, TEXT, u'FOO'),
             (None, END, u'foo'),
             (None, START, u'bar'),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_append_context(self):
        self.assertEqual(
            self._apply('.'),
            [(ENTER, START, u'root'),
             (INSIDE, TEXT, u'ROOT'),
             (INSIDE, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (INSIDE, END, u'foo'),
             (INSIDE, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (INSIDE, END, u'bar'),
             (None, TEXT, u'CONTENT 1'),
             (EXIT, END, u'root')],
            )

    def test_append_text_context(self):
        self.assertEqual(
            self._apply('.', html='foo'),
            [(OUTSIDE, TEXT, u'foo')]
            )

    def test_append_adjacent_elements(self):
        self.assertEqual(
            self._apply('*'),
            [(None, START, u'root'),
             (None, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (None, TEXT, u'CONTENT 1'),
             (EXIT, END, u'foo'),
             (ENTER, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (None, TEXT, u'CONTENT 2'),
             (EXIT, END, u'bar'),
             (None, END, u'root')]

            )

    def test_append_all(self):
        self.assertEqual(
            self._apply('*|text()'),
            [(None, START, u'root'),
             (OUTSIDE, TEXT, u'ROOT'),
             (ENTER, START, u'foo'),
             (INSIDE, TEXT, u'FOO'),
             (None, TEXT, u'CONTENT 1'),
             (EXIT, END, u'foo'),
             (ENTER, START, u'bar'),
             (INSIDE, TEXT, u'BAR'),
             (None, TEXT, u'CONTENT 2'),
             (EXIT, END, u'bar'),
             (None, END, u'root')]
            )

    def test_append_with_callback(self):
        count = [0]
        def content():
            count[0] += 1
            yield '%2i.' % count[0]
        self.assertEqual(
            self._apply('foo', content),
            [(None, 'START', u'root'),
             (None, 'TEXT', u'ROOT'),
             (ENTER, 'START', u'foo'),
             (INSIDE, 'TEXT', u'FOO'),
             (None, 'TEXT', u' 1.'),
             (EXIT, 'END', u'foo'),
             (None, 'START', u'bar'),
             (None, 'TEXT', u'BAR'),
             (None, 'END', u'bar'),
             (None, 'END', u'root')]
            )



class AttrTest(unittest.TestCase):
    def _attr(self, select, name, value):
        return _transform(FOOBAR, Transformer(select).attr(name, value),
                          with_attrs=True)

    def test_set_existing_attr(self):
        self.assertEqual(
            self._attr('foo', 'name', 'FOO'),
            [(None, START, (u'root', {})),
             (None, TEXT, u'ROOT'),
             (ENTER, START, (u'foo', {u'name': 'FOO', u'size': '100'})),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, START, (u'bar', {u'name': u'bar'})),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_set_new_attr(self):
        self.assertEqual(
            self._attr('foo', 'title', 'FOO'),
            [(None, START, (u'root', {})),
             (None, TEXT, u'ROOT'),
             (ENTER, START, (u'foo', {u'name': u'foo', u'title': 'FOO', u'size': '100'})),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, START, (u'bar', {u'name': u'bar'})),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_attr_from_function(self):
        def set(name, event):
            self.assertEqual(name, 'name')
            return event[1][1].get('name').upper()

        self.assertEqual(
            self._attr('foo|bar', 'name', set),
            [(None, START, (u'root', {})),
             (None, TEXT, u'ROOT'),
             (ENTER, START, (u'foo', {u'name': 'FOO', u'size': '100'})),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (ENTER, START, (u'bar', {u'name': 'BAR'})),
             (INSIDE, TEXT, u'BAR'),
             (EXIT, END, u'bar'),
             (None, END, u'root')]
            )

    def test_remove_attr(self):
        self.assertEqual(
            self._attr('foo', 'name', None),
            [(None, START, (u'root', {})),
             (None, TEXT, u'ROOT'),
             (ENTER, START, (u'foo', {u'size': '100'})),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, START, (u'bar', {u'name': u'bar'})),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )

    def test_remove_attr_with_function(self):
        def set(name, event):
            return None

        self.assertEqual(
            self._attr('foo', 'name', set),
            [(None, START, (u'root', {})),
             (None, TEXT, u'ROOT'),
             (ENTER, START, (u'foo', {u'size': '100'})),
             (INSIDE, TEXT, u'FOO'),
             (EXIT, END, u'foo'),
             (None, START, (u'bar', {u'name': u'bar'})),
             (None, TEXT, u'BAR'),
             (None, END, u'bar'),
             (None, END, u'root')]
            )


class BufferTestMixin(object):
    def _apply(self, select, with_attrs=False):
        buffer = StreamBuffer()
        events = buffer.events

        class Trace(object):
            last = None
            trace = []

            def __call__(self, stream):
                for event in stream:
                    if events and hash(tuple(events)) != self.last:
                        self.last = hash(tuple(events))
                        self.trace.append(list(events))
                    yield event

        trace = Trace()
        output = _transform(FOOBAR, getattr(Transformer(select), self.operation)
                                    (buffer).apply(trace), with_attrs=with_attrs)
        simplified = []
        for interval in trace.trace:
            simplified.append(_simplify([(None, e) for e in interval],
                                         with_attrs=with_attrs))
        return output, simplified


class CopyTest(unittest.TestCase, BufferTestMixin):
    operation = 'copy'

    def test_copy_element(self):
        self.assertEqual(
            self._apply('foo')[1],
            [[(None, START, u'foo'),
              (None, TEXT, u'FOO'),
              (None, END, u'foo')]]
            )

    def test_copy_adjacent_elements(self):
        self.assertEqual(
            self._apply('foo|bar')[1],
            [[(None, START, u'foo'),
              (None, TEXT, u'FOO'),
              (None, END, u'foo')],
             [(None, START, u'bar'),
              (None, TEXT, u'BAR'),
              (None, END, u'bar')]]
            )

    def test_copy_all(self):
        self.assertEqual(
            self._apply('*|text()')[1],
            [[(None, TEXT, u'ROOT')],
             [(None, START, u'foo'),
              (None, TEXT, u'FOO'),
              (None, END, u'foo')],
             [(None, START, u'bar'),
              (None, TEXT, u'BAR'),
              (None, END, u'bar')]]
            )

    def test_copy_text(self):
        self.assertEqual(
            self._apply('*/text()')[1],
            [[(None, TEXT, u'FOO')],
             [(None, TEXT, u'BAR')]]
            )

    def test_copy_context(self):
        self.assertEqual(
            self._apply('.')[1],
            [[(None, START, u'root'),
              (None, TEXT, u'ROOT'),
              (None, START, u'foo'),
              (None, TEXT, u'FOO'),
              (None, END, u'foo'),
              (None, START, u'bar'),
              (None, TEXT, u'BAR'),
              (None, END, u'bar'),
              (None, END, u'root')]]
            )

    def test_copy_attribute(self):
        self.assertEqual(
            self._apply('foo/@name', with_attrs=True)[1],
            [[(None, ATTR, {'name': u'foo'})]]
            )

    def test_copy_attributes(self):
        self.assertEqual(
            self._apply('foo/@*', with_attrs=True)[1],
            [[(None, ATTR, {u'name': u'foo', u'size': u'100'})]]
            )


class CutTest(unittest.TestCase, BufferTestMixin):
    operation = 'cut'

    def test_cut_element(self):
        self.assertEqual(
            self._apply('foo'),
            ([(None, START, u'root'),
              (None, TEXT, u'ROOT'),
              (None, START, u'bar'),
              (None, TEXT, u'BAR'),
              (None, END, u'bar'),
              (None, END, u'root')],
             [[(None, START, u'foo'),
               (None, TEXT, u'FOO'),
               (None, END, u'foo')]])
            )

    def test_cut_adjacent_elements(self):
        self.assertEqual(
            self._apply('foo|bar'),
            ([(None, START, u'root'), 
              (None, TEXT, u'ROOT'),
              (BREAK, BREAK, None),
              (None, END, u'root')],
             [[(None, START, u'foo'),
               (None, TEXT, u'FOO'),
               (None, END, u'foo')],
              [(None, START, u'bar'),
               (None, TEXT, u'BAR'),
               (None, END, u'bar')]])
            )

    def test_cut_all(self):
        self.assertEqual(
            self._apply('*|text()'),
            ([(None, 'START', u'root'),
              ('BREAK', 'BREAK', None),
              ('BREAK', 'BREAK', None),
              (None, 'END', u'root')],
             [[(None, 'TEXT', u'ROOT')],
              [(None, 'START', u'foo'),
               (None, 'TEXT', u'FOO'),
               (None, 'END', u'foo')],
              [(None, 'START', u'bar'),
               (None, 'TEXT', u'BAR'),
               (None, 'END', u'bar')]])
            )

    def test_cut_text(self):
        self.assertEqual(
            self._apply('*/text()'),
            ([(None, 'START', u'root'),
              (None, 'TEXT', u'ROOT'),
              (None, 'START', u'foo'),
              (None, 'END', u'foo'),
              (None, 'START', u'bar'),
              (None, 'END', u'bar'),
              (None, 'END', u'root')],
             [[(None, 'TEXT', u'FOO')],
              [(None, 'TEXT', u'BAR')]])
            )

    def test_cut_context(self):
        self.assertEqual(
            self._apply('.')[1],
            [[(None, 'START', u'root'),
              (None, 'TEXT', u'ROOT'),
              (None, 'START', u'foo'),
              (None, 'TEXT', u'FOO'),
              (None, 'END', u'foo'),
              (None, 'START', u'bar'),
              (None, 'TEXT', u'BAR'),
              (None, 'END', u'bar'),
              (None, 'END', u'root')]]
            )

    def test_cut_attribute(self):
        self.assertEqual(
            self._apply('foo/@name', with_attrs=True),
            ([(None, START, (u'root', {})),
              (None, TEXT, u'ROOT'),
              (None, START, (u'foo', {u'size': u'100'})),
              (None, TEXT, u'FOO'),
              (None, END, u'foo'),
              (None, START, (u'bar', {u'name': u'bar'})),
              (None, TEXT, u'BAR'),
              (None, END, u'bar'),
              (None, END, u'root')],
             [[(None, ATTR, {u'name': u'foo'})]])
            )

    def test_cut_attributes(self):
        self.assertEqual(
            self._apply('foo/@*', with_attrs=True),
            ([(None, START, (u'root', {})),
              (None, TEXT, u'ROOT'),
              (None, START, (u'foo', {})),
              (None, TEXT, u'FOO'),
              (None, END, u'foo'),
              (None, START, (u'bar', {u'name': u'bar'})),
              (None, TEXT, u'BAR'),
              (None, END, u'bar'),
              (None, END, u'root')],
             [[(None, ATTR, {u'name': u'foo', u'size': u'100'})]])
            )

# XXX Test this when the XPath implementation is fixed (#233).
#    def test_cut_attribute_or_attribute(self):
#        self.assertEqual(
#            self._apply('foo/@name | foo/@size', with_attrs=True),
#            ([(None, START, (u'root', {})),
#              (None, TEXT, u'ROOT'),
#              (None, START, (u'foo', {})),
#              (None, TEXT, u'FOO'),
#              (None, END, u'foo'),
#              (None, START, (u'bar', {u'name': u'bar'})),
#              (None, TEXT, u'BAR'),
#              (None, END, u'bar'),
#              (None, END, u'root')],
#             [[(None, ATTR, {u'name': u'foo', u'size': u'100'})]])
#            )




def suite():
    from genshi.input import HTML
    from genshi.core import Markup
    from genshi.builder import tag
    suite = unittest.TestSuite()
    for test in (SelectTest, InvertTest, EndTest,
                 EmptyTest, RemoveTest, UnwrapText, WrapTest, FilterTest,
                 MapTest, SubstituteTest, RenameTest, ReplaceTest, BeforeTest,
                 AfterTest, PrependTest, AppendTest, AttrTest, CopyTest, CutTest):
        suite.addTest(unittest.makeSuite(test, 'test'))
    suite.addTest(doctest.DocTestSuite(
        genshi.filters.transform, optionflags=doctest.NORMALIZE_WHITESPACE,
        extraglobs={'HTML': HTML, 'tag': tag, 'Markup': Markup}))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')
